//================================================================
//
//#	Copyright 	Earth People Technology Inc. 2016
//
//
//	File Name:	i2c_comms_interface.v
//	Author:	    R. Jolly
//	Date:		July 26, 2016
//	Revision:	A
//
//
//# Development: EPT I2C Master Project 
//# Application: Altera Cyclone IV FPGA
//# Description: This file contains verilog code which will allow access
//#              to Active Transfer Library. 
//				
//
//
//
//	Revision History:
//
//	DATE		VERSION		DETAILS
//	7/26/16 		A		Created by R Jolly
//  8/22/19         B       Changed I2C fixed parameters to flexible
//                          generics in uc_controller_define.v
//
//
//
//
//
//  Register Mapping:
//           STATUS_REGISTER[0] = Read a Receive word from the I2C Unit
//           STATUS_REGISTER[1] = Write a Transmit word to the I2C Unit
//           STATUS_REGISTER[2] = Latch the Address into the I2C Unit
//           STATUS_REGISTER[3] = Request Receive words from the I2C Unit
//           STATUS_REGISTER[4] = 
//           STATUS_REGISTER[5] = 
//           STATUS_REGISTER[6] = 
//           STATUS_REGISTER[7] = 
//
//
//
//
//
//==================================================================

`resetall
`timescale 1ns/10ps

`include "../src/uc_controller_define.v"



module i2c_comms_interface
(
	input  wire                 CLK,
	input  wire                 RST_N,

	input  wire                 SERIAL_CLK,

	inout  wire                 I2C_DATA,
	output wire                 I2C_CLOCK,

	inout   wire [7:0]          MEMORY_ADDRESS,
	inout   wire [7:0]          MEMORY_DATA,
	output  wire                MEMORY_WRITE_LATCH,
	output  wire                MEMORY_READ_LATCH,
	output  wire                MEMORY_OUTPUT_ENABLE,
	
	input   wire [7:0]          I2C_COMMAND,
	input   wire                I2C_COMMAND_LATCH,
	
	output  reg                 I2C_TRANSMIT_COMPLETE,
	output  wire                I2C_BYTE_RECEIVED,

	output  reg                 REQUEST_EN,
	input   wire                GRANT_RDY
	
		);

//==================================================================
// Parameter Declarations
//==================================================================
    parameter   IDLE                       = 0,
                START                      = 1,
                LATCH_COMMAND              = 2,
                CHECK_READ_WRITE_FLAG      = 3,
                MEMORY_RD_ASSERT_OE        = 4,
                MEMORY_RD_ADDRESS          = 5,
                MEMORY_RD_DATA             = 6,
                MEMORY_RD_LATCH            = 7,
				MEMORY_RD_CHECK            = 8,
				MEMORY_RD_DELAY            = 9,
                START_I2C_TRANSMIT         = 10,
				WAIT_FOR_I2C_COMPLETE      = 11,
                CHECK_I2C_COMPLETE         = 12,
				START_I2C_TRANSMIT_DELAY   = 13,
                START_I2C_RECEIVE          = 14,
				COMPLETE_I2C_RECEIVE       = 15,
                MEMORY_WRITE_CYCLE         = 16,
                MEMORY_WR_ADDRESS          = 17,
                MEMORY_WR_DATA             = 18,
                MEMORY_WR_LATCH            = 19;

//==================================================================
// IO Port Declarations
//==================================================================

//==================================================================
// Register Declarations
//==================================================================
  
	// State Machine Registers
	reg	[19:0]                  state;
	reg	[19:0]                  next;

	//Receive Storage registers
	reg  [7:0]                  serial_receive_data_storage;	
    reg  [1:0]                  serial_receive_data_reg;
	wire [7:0]                  transfer_1_address;

	//Serial I2C Transmit Registers
	reg    [7:0]                i2c_transfer_storage_register;
	wire                        serial_transmit_busy;
	wire                        master_i2c_transmit_complete;
	reg                         master_i2c_initiate_receive;
	wire                        master_i2c_byte_received;
	wire   [7:0]                serial_receive_out_byte;
	reg                         master_i2c_initiate_transmit;
	reg    [7:0]                i2c_command_to_send;
	reg    [3:0]                number_of_i2c_read_bytes;
	reg                         i2c_read_write_flag;
	reg    [7:0]                i2c_payload_bytes [3:0];
	reg                         i2c_payload_byte_count_reg;
	wire                        i2c_cs_from_i2c_iface;
	wire                        master_i2c_transmit_byte_done;
	wire                        master_i2c_receive_done;
	reg                         master_i2c_read_byte;

	//I2C Transmit Byte counters
	reg    [3:0]                number_of_i2c_payload_bytes;
	reg    [3:0]                number_of_i2c_bytes_to_send;
	reg    [3:0]                i2c_payload_byte_count;	
	reg    [3:0]                i2c_transmit_total_count;
	reg                         i2c_transmit_total_count_reg;

	
	//I2C Receive Registers
	reg    [3:0]                i2c_receive_total_count;
	reg                         start_i2c_receive_reg;
	
	
	// Read Cycle Complete delay counter
	//reg [7:0]                   read_complete_counter;
	//reg                         read_complete_flag;
	
	//Memory Read registers
	reg                         memory_rd_address_flag;
	reg                         memory_rd_latch_flag;
	reg                         memory_rd_data_flag;
	reg                         memory_output_enable_flag;	
	reg                         memory_rd_latch_reg;
	reg [7:0]                   memory_rd_delay_counter;
	
	//Memory Write registers
	reg                         memory_wr_address_flag;
	reg                         memory_wr_latch_reg;
	reg                         memory_wr_data_flag;
	reg                         memory_wr_latch_flag;
	
	//Memory Interface registers
	reg [7:0]                   i2c_payload_start_address_in_memory;
	
	//I2C Command Storage from UC Controller
	reg [7:0]                   i2c_command_storage;	
	

//==================================================================
// Simulation Directives
//==================================================================

`ifdef SIM
	reg [8*30:1] state_name;
`endif

//=================================================================================
// Assignments	
//=================================================================================

    //Memory Bus Controller
	assign          MEMORY_ADDRESS = (memory_wr_address_flag | memory_rd_address_flag) ? i2c_payload_start_address_in_memory: 8'hzz;
	assign          MEMORY_DATA = memory_wr_data_flag ? serial_receive_out_byte : 8'hzz;
	assign          MEMORY_WRITE_LATCH = memory_wr_latch_flag ? 1'b1 : 1'b0;
	assign          MEMORY_READ_LATCH = memory_rd_latch_flag ? 1'b1 : 1'b0;
	assign          MEMORY_OUTPUT_ENABLE = memory_output_enable_flag ? 1'b1 : 1'b0;
	
	assign          I2C_BYTE_RECEIVED = master_i2c_byte_received;


//=================================================================================
//  I2C Command Cycle: Read in the command from the uc controller
//=================================================================================

	always @( negedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		begin
		   i2c_command_storage <= 0;
		end
  		else
    	begin
		   if(I2C_COMMAND_LATCH)
		      i2c_command_storage <= I2C_COMMAND;
    	end
  	end

//=================================================================================
//  I2C Command Cycle: Latch the I2C Command to send over I2C
//=================================================================================

	always @( negedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		begin
			number_of_i2c_bytes_to_send <= 0;
			i2c_command_to_send <= 0;
			i2c_read_write_flag <= 1'b0;
			number_of_i2c_payload_bytes <= 0;
			number_of_i2c_read_bytes <= 3'h1;
			i2c_payload_start_address_in_memory <= 0;
		    memory_wr_latch_reg <= 1'b0;
			memory_rd_latch_reg <= 1'b0;
		end
  		else
    	begin
		   if(state[LATCH_COMMAND])
		   begin
		      case(i2c_command_storage[7:4])
			  `TMP102_WRITE_TO_TEMP_REG: //Write to TMP102 Temperature Register 2'b00
			  begin
			     number_of_i2c_bytes_to_send <= 3'h2;
				 number_of_i2c_payload_bytes <= 3'h1;
				 i2c_command_to_send <= `TMP102_FIXED_ADDRESS;
				 i2c_command_to_send[0] <= `I2C_WRITE_COMMAND;
				 i2c_read_write_flag <= 1'b0;
				 i2c_payload_start_address_in_memory <= `I2C_0_WRITE_A_CMD_MEMORY_LOCATION;
			  end
			  `TMP102_WRITE_TO_CONFIG_REG: //Write to TMP102 Configuration Register 2'b01
			  begin
			     number_of_i2c_bytes_to_send <= 3'h2;
				 number_of_i2c_payload_bytes <= 3'h1;
				 i2c_command_to_send <= `TMP102_FIXED_ADDRESS;
				 i2c_command_to_send[0] <= `I2C_WRITE_COMMAND;
				 i2c_read_write_flag <= 1'b0;
				 i2c_payload_start_address_in_memory <= `I2C_1_WRITE_B_CMD_MEMORY_LOCATION;
			  end
			  `TMP102_READ_FROM_TEMP_REG: //Read from TMP102 Temperature Register
			  begin
			     number_of_i2c_bytes_to_send <= 3'h2;
				 number_of_i2c_payload_bytes <= 3'h2;
				 i2c_command_to_send <= `TMP102_FIXED_ADDRESS;
				 i2c_command_to_send[0] <= `I2C_READ_COMMAND;
				 i2c_read_write_flag <= 1'b1;
				 i2c_payload_start_address_in_memory <= `I2C_0_READ_A_DATA1_MEMORY_LOCATION;
			  end
			  `TMP102_READ_FROM_CONFIG_REG: //Read from TMP102 Configuration Register 2'b01
			  begin
				 number_of_i2c_payload_bytes <= 3'h2;
				 number_of_i2c_bytes_to_send <= 3'h2;
				 number_of_i2c_read_bytes <= 3'h2;
				 i2c_command_to_send <= `TMP102_FIXED_ADDRESS;
				 i2c_command_to_send[0] <= `I2C_READ_COMMAND;
				 i2c_read_write_flag <= 1'b1;
				 i2c_payload_start_address_in_memory <= `I2C_1_READ_B_DATA1_MEMORY_LOCATION;
			  end
			  endcase
		   end
           else if(state[CHECK_I2C_COMPLETE] & !MEMORY_WRITE_LATCH )
		   begin
		      memory_wr_latch_reg <= 1'b0;
		      //if(number_of_i2c_bytes_to_send > 0)
		      //   number_of_i2c_bytes_to_send <= number_of_i2c_bytes_to_send - 1'd1;
			  //if(number_of_i2c_payload_bytes > 0)
              //   number_of_i2c_payload_bytes <= number_of_i2c_payload_bytes - 1'd1;
		   end
           else if(state[CHECK_I2C_COMPLETE] & MEMORY_WRITE_LATCH & !memory_wr_latch_reg)
		   begin
		      memory_wr_latch_reg <= 1'b1;
		      i2c_payload_start_address_in_memory <= i2c_payload_start_address_in_memory + 1'd1;
		   end
           else if(state[MEMORY_RD_CHECK] & !memory_rd_latch_reg)
		   begin
		      memory_rd_latch_reg <= 1'b1;
		      i2c_payload_start_address_in_memory <= i2c_payload_start_address_in_memory + 1'd1;
		   end
           else if(state[IDLE] | (state[MEMORY_WRITE_CYCLE] & !MEMORY_WRITE_LATCH & memory_wr_latch_reg) | state[MEMORY_RD_DELAY])
		   begin
		      memory_wr_latch_reg <= 1'b0;
			  memory_rd_latch_reg <= 1'b0;
		   end
   	end
  	end

  
//=================================================================================
//  Memory Access Cycle: REQUEST_EN
//=================================================================================

	always @( negedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		begin
		   REQUEST_EN <= 1'b0;
		end
  		else
    	begin
		   if(state[CHECK_READ_WRITE_FLAG] & !GRANT_RDY)
		      REQUEST_EN <= 1'b1;
		   else if(state[IDLE])//state[MEMORY_RD_ADDRESS] | state[MEMORY_WR_ADDRESS])
		     REQUEST_EN <= 1'b0;
    	end
  	end

		
//=================================================================================
//	I2C Receive Cycle: Start 
//=================================================================================
	always @( posedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		begin
		   master_i2c_initiate_receive <= 1'b0;
		   i2c_receive_total_count <= 0;
		   start_i2c_receive_reg <= 1'b0;
		end
		else
		begin
		   if(state[START_I2C_TRANSMIT])
		   begin
		      i2c_receive_total_count <= number_of_i2c_read_bytes;
		   end
		   else if(state[START_I2C_RECEIVE] & !start_i2c_receive_reg)
		   begin
		      start_i2c_receive_reg <= 1'b1;
		      if(i2c_receive_total_count > 0)
		         i2c_receive_total_count <= i2c_receive_total_count - 1'd1;
		      master_i2c_initiate_receive <= 1'b1;
		   end
		   else if(state[MEMORY_WR_ADDRESS])
		   begin
		      start_i2c_receive_reg <= 1'b0;
		      master_i2c_initiate_receive <= 1'b0;
		   end
		end
	end

//=================================================================================
//  Read Cycle: Assert the I2C_RECEIVE_ENA_IN
//=================================================================================

	always @( negedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		begin
		   master_i2c_read_byte <= 1'b0;
		end
  		else
		   if(master_i2c_byte_received & master_i2c_receive_done)
		      master_i2c_read_byte <= 1'b1;
		   else if(state[MEMORY_WR_LATCH])
		      master_i2c_read_byte <= 1'b0;
   	end

  
//=================================================================================
//  Read Cycle: Memory Output Enable
//=================================================================================

	always @( negedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		begin
		   memory_output_enable_flag <= 1'b0;
		end
  		else
		   if(state[MEMORY_RD_ASSERT_OE])
		      memory_output_enable_flag <= 1'b1;
		   else if(state[MEMORY_RD_LATCH])
		      memory_output_enable_flag <= 1'b0;
   	end

//=================================================================================
//	Read Cycle: Memory Read Address
//=================================================================================
	always @( posedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		   memory_rd_data_flag <= 1'b0;
		else
		begin
		   if(state[MEMORY_RD_DATA])
		      memory_rd_data_flag <= 1'b1;
		   else if(state[CHECK_I2C_COMPLETE])
		      memory_rd_data_flag <= 1'b0;
		end
	end

//=================================================================================
//	Read Cycle: Memory Read Data
//=================================================================================
	always @( posedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		   memory_rd_address_flag <= 1'b0;
		else
		begin
		   if(state[MEMORY_RD_ASSERT_OE])
		   begin
		      memory_rd_address_flag <= 1'b1;
		   end
		   else if(state[MEMORY_RD_CHECK])
		      memory_rd_address_flag <= 1'b0;
		end
	end
	
//=================================================================================
//	Read Cycle: Memory Read Latch
//=================================================================================
	always @( posedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		   memory_rd_latch_flag <= 1'b0;
		else
		begin
		   if(state[MEMORY_RD_LATCH])
		   begin
		      memory_rd_latch_flag <= 1'b1;
		   end
		   else if(state[MEMORY_RD_CHECK])
		      memory_rd_latch_flag <= 1'b0;
		end
	end
			
//=================================================================================
//	Read Cycle: Memory Read Wait States
//=================================================================================
	always @( posedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		   memory_rd_delay_counter <= 0;
		else
		begin
		   if(state[MEMORY_RD_DELAY] | state[MEMORY_RD_LATCH])
		   begin
		      memory_rd_delay_counter <= memory_rd_delay_counter + 1'd1;
		   end
		   else if(state[MEMORY_RD_ASSERT_OE] | state[MEMORY_RD_CHECK])
		      memory_rd_delay_counter <= 0;
		end
	end
			
//=================================================================================
//	I2C Transmit Cycle: Start 
//=================================================================================
	always @( posedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		begin
		   master_i2c_initiate_transmit <= 1'b0;
		   i2c_transmit_total_count <= 0;
		   i2c_transmit_total_count_reg <= 1'b0;
		end
		else
		begin
		   if(state[START_I2C_TRANSMIT] & !i2c_read_write_flag)
		   begin
		      master_i2c_initiate_transmit <= 1'b1;
			  i2c_transfer_storage_register <= i2c_payload_bytes[i2c_transmit_total_count];
		   end
		   else if(state[WAIT_FOR_I2C_COMPLETE] & !i2c_transmit_total_count_reg  & !i2c_read_write_flag)
		   begin
		      i2c_transmit_total_count <= i2c_transmit_total_count + 1'd1;
		      master_i2c_initiate_transmit <= 1'b0;
			  i2c_transmit_total_count_reg <= 1'b1;
		   end
		   else if(state[CHECK_I2C_COMPLETE] & i2c_transmit_total_count_reg  & !i2c_read_write_flag)
		   begin
			  i2c_transmit_total_count_reg <= 1'b0;
		   end
		   else if(state[START_I2C_TRANSMIT] & i2c_read_write_flag)
		   begin
		      master_i2c_initiate_transmit <= 1'b1;
			  i2c_transfer_storage_register <= i2c_payload_bytes[0];
		   end
		   else if(state[WAIT_FOR_I2C_COMPLETE] & !i2c_transmit_total_count_reg  & i2c_read_write_flag)
		   begin
		      i2c_transmit_total_count <= 0;
		      master_i2c_initiate_transmit <= 1'b0;
			  i2c_transmit_total_count_reg <= 1'b1;
		   end
		   else if(state[CHECK_I2C_COMPLETE] & i2c_transmit_total_count_reg  & i2c_read_write_flag)
		   begin
			  i2c_transmit_total_count_reg <= 1'b0;
		   end
		   else if(state[IDLE])
		   begin
		      master_i2c_initiate_transmit <= 1'b0;
		      i2c_transmit_total_count <= 0;
			  i2c_transmit_total_count_reg <= 1'b0;
		   end
		end
	end

//=================================================================================
//	I2C Transmit Cycle: Load I2C Transmit Array 
//=================================================================================
	always @( posedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		begin
		   i2c_payload_byte_count <= 0;
		   i2c_payload_byte_count_reg <= 1'b0;
		end
		else
		begin
		   if(state[LATCH_COMMAND])
		   begin
			  i2c_payload_bytes[i2c_payload_byte_count] <= i2c_command_to_send;
		   end
		   else if(state[CHECK_READ_WRITE_FLAG] & i2c_read_write_flag)
		   begin
			  //Special cases to handle PGA_READ_COMMAND
			     i2c_payload_byte_count <= i2c_payload_byte_count + 1'd1;
			     i2c_payload_bytes[1] <= 8'h00;
		   end
		   else if(state[MEMORY_RD_ASSERT_OE] & !i2c_payload_byte_count_reg)
		   begin
		      i2c_payload_byte_count_reg <= 1'b1;
		      i2c_payload_byte_count <= i2c_payload_byte_count + 1'd1;
		   end
		   else if(state[MEMORY_RD_CHECK])
		   begin
		      i2c_payload_byte_count_reg <= 1'b0;
			  i2c_payload_bytes[i2c_payload_byte_count] <= MEMORY_DATA;
		   end
		   else if(state[IDLE])
		   begin
		      i2c_payload_byte_count <= 0;
		      i2c_payload_byte_count_reg <= 1'b0;
		   end
		end
	end
			
//=================================================================================
//  I2C Transmit Cycle: Transmission Complete
//=================================================================================

	always @( negedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		begin
		   I2C_TRANSMIT_COMPLETE <= 1'b0;
		end
  		else
		begin
		   if(state[CHECK_I2C_COMPLETE] & (i2c_transmit_total_count >= number_of_i2c_bytes_to_send))
		      I2C_TRANSMIT_COMPLETE <= 1'b1;
		   else if(state[IDLE])
		      I2C_TRANSMIT_COMPLETE <= 1'b0;
		end
   	end

	
//=================================================================================
//	Write Cycle: Memory Write Address
//=================================================================================
	always @( posedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		   memory_wr_address_flag <= 1'b0;
		else
		begin
		   if(state[MEMORY_WR_ADDRESS])
		      memory_wr_address_flag <= 1'b1;
		   else if(state[CHECK_I2C_COMPLETE])
		      memory_wr_address_flag <= 1'b0;
		end
	end

//=================================================================================
//	Write Cycle: Memory Write Data
//=================================================================================
	always @( posedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		   memory_wr_data_flag <= 1'b0;
		else
		begin
		   if(state[MEMORY_WR_DATA])
		   begin
		      memory_wr_data_flag <= 1'b1;
		   end
		   else if(state[CHECK_I2C_COMPLETE])
		      memory_wr_data_flag <= 1'b0;
		end
	end
	
//=================================================================================
//	Write Cycle: Memory Write Latch
//=================================================================================
	always @( posedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		   memory_wr_latch_flag <= 1'b0;
		else
		begin
		   if(state[MEMORY_WR_LATCH])
		   begin
		      memory_wr_latch_flag <= 1'b1;
		   end
		   else if(state[CHECK_I2C_COMPLETE])
		      memory_wr_latch_flag <= 1'b0;
		end
	end
			

//=================================================================================
//	Serial I2C Communications Finite State Machine
//              IDLE                       = 0,
//              START                      = 1,
//              LATCH_COMMAND              = 2,
//              CHECK_READ_WRITE_FLAG      = 3,
//              MEMORY_RD_ASSERT_OE        = 4,
//              MEMORY_RD_ADDRESS          = 5,
//              MEMORY_RD_DATA             = 6,
//              MEMORY_RD_LATCH            = 7,
//              START_I2C_TRANSMIT         = 8,
//              CHECK_I2C_COMPLETE         = 9,
//              START_I2C_TRANSMIT_DELAY   = 10
//              MEMORY_WRITE_CYCLE         = 11,
//              MEMORY_WR_ADDRESS          = 12,
//              MEMORY_WR_DATA             = 13,
//              MEMORY_WR_LATCH            = 14;
//
//=================================================================================

	always @(  posedge CLK or negedge RST_N  )
	begin
		if( !RST_N )
		begin
			state <= 20'h00;
			state [IDLE] <= 1'b1;
		end
		else
			state <= next;
	end


	always @( state or I2C_COMMAND_LATCH or MEMORY_READ_LATCH or i2c_read_write_flag
	         or GRANT_RDY or i2c_payload_byte_count or master_i2c_transmit_complete
			 or number_of_i2c_bytes_to_send or master_i2c_byte_received
			 or memory_rd_latch_flag or master_i2c_initiate_receive
			 or memory_rd_delay_counter or master_i2c_transmit_byte_done
			 or master_i2c_read_byte or master_i2c_receive_done)
	begin
		next = 20'h00;
		if(state[IDLE])
		begin
			if(I2C_COMMAND_LATCH)
				next[START] = 1'b1;
			else
				next[IDLE] = 1'b1;
		end

		if(state[START])
		begin
          next[LATCH_COMMAND] = 1'b1;
		end

		if(state[LATCH_COMMAND])
		begin
			next[CHECK_READ_WRITE_FLAG] = 1'b1;
		end

		if(state[CHECK_READ_WRITE_FLAG])
		begin
		    if(!REQUEST_EN)
				next[CHECK_READ_WRITE_FLAG] = 1'b1;
			else if(!i2c_read_write_flag)
				next[MEMORY_RD_ASSERT_OE] = 1'b1;
			else
				next[START_I2C_TRANSMIT] = 1'b1;
		end

		if(state[MEMORY_RD_ASSERT_OE])
		begin
			if(GRANT_RDY & !MEMORY_READ_LATCH)
				next[MEMORY_RD_ADDRESS] = 1'b1;
			else
				next[MEMORY_RD_ASSERT_OE] = 1'b1;
		end

		if(state[MEMORY_RD_ADDRESS])
		begin
           next[MEMORY_RD_DATA] = 1'b1;
		end

		if(state[MEMORY_RD_DATA])
		begin
           next[MEMORY_RD_LATCH] = 1'b1;
		end

		if(state[MEMORY_RD_LATCH])
     	begin
			if(memory_rd_delay_counter > `MEMORY_RD_WAIT_STATES)
				next[MEMORY_RD_CHECK] = 1'b1;
			else
		       next[MEMORY_RD_LATCH] = 1'b1;
		end

		if(state[MEMORY_RD_CHECK])
		begin
		   if(i2c_payload_byte_count < /*number_of_i2c_bytes_to_send*/number_of_i2c_payload_bytes)
		      next[MEMORY_RD_DELAY] = 1'b1;
		   else if(memory_rd_latch_flag)
		      next[START_I2C_TRANSMIT] = 1'b1;
		   else
		      next[MEMORY_RD_CHECK] = 1'b1;
		end

		if(state[MEMORY_RD_DELAY])
		begin
			if(memory_rd_delay_counter > `MEMORY_RD_WAIT_STATES)
				next[MEMORY_RD_ASSERT_OE] = 1'b1;
			else
		   next[MEMORY_RD_DELAY] = 1'b1;
		end

		if(state[START_I2C_TRANSMIT])
		begin
				next[WAIT_FOR_I2C_COMPLETE] = 1'b1;
		end

		if(state[WAIT_FOR_I2C_COMPLETE])
		begin
			if(!i2c_read_write_flag & master_i2c_transmit_complete)
				next[CHECK_I2C_COMPLETE] = 1'b1;
			else if(i2c_read_write_flag & master_i2c_transmit_byte_done)
			    next[CHECK_I2C_COMPLETE] = 1'b1;
			else
				next[WAIT_FOR_I2C_COMPLETE] = 1'b1;
		end

		if(state[CHECK_I2C_COMPLETE])
		begin
			if(i2c_read_write_flag)
			begin
			   if(i2c_receive_total_count >= 3'h1)
			      next[START_I2C_RECEIVE] = 1'b1;
			   else
                  next[IDLE] = 1'b1;
			end
            else if(i2c_transmit_total_count < number_of_i2c_bytes_to_send)
				next[START_I2C_TRANSMIT_DELAY] = 1'b1;
			else
				next[IDLE] = 1'b1;
		end

		if(state[START_I2C_TRANSMIT_DELAY])
		begin
			if(!master_i2c_transmit_complete)
				next[START_I2C_TRANSMIT] = 1'b1;
			else
				next[START_I2C_TRANSMIT_DELAY] = 1'b1;
		end

		if(state[START_I2C_RECEIVE])
		begin
			if(master_i2c_initiate_receive)
				next[COMPLETE_I2C_RECEIVE] = 1'b1;
			else
				next[START_I2C_RECEIVE] = 1'b1;
		end

		if(state[COMPLETE_I2C_RECEIVE])
		begin
			if(master_i2c_receive_done)
				next[MEMORY_WRITE_CYCLE] = 1'b1;
			else
				next[COMPLETE_I2C_RECEIVE] = 1'b1;
		end

		if(state[MEMORY_WRITE_CYCLE])
		begin
			if(master_i2c_read_byte)
			   next[MEMORY_WR_ADDRESS] = 1'b1;
			else
			   next[MEMORY_WRITE_CYCLE] = 1'b1;
		end

		if(state[MEMORY_WR_ADDRESS])
		begin
           next[MEMORY_WR_DATA] = 1'b1;
		end

		if (state[MEMORY_WR_DATA])
		begin 
           next[MEMORY_WR_LATCH] = 1'b1;
		end

		if (state[MEMORY_WR_LATCH])
		//begin
	    //   if(master_i2c_byte_received)
              next[CHECK_I2C_COMPLETE] = 1'b1;
		//	else
        //       next[MEMORY_WR_LATCH] = 1'b1;
		//end

`ifdef SIM
begin
	if ( state == (1 << IDLE))
		state_name = "IDLE";		
	if ( state == (1 << START))
		state_name = "START";		
	if ( state == (1 << LATCH_COMMAND))
		state_name = "LATCH_COMMAND";		
	if ( state == (1 << CHECK_READ_WRITE_FLAG))
		state_name = "CHECK_READ_WRITE_FLAG";		
	if ( state == (1 << MEMORY_RD_ASSERT_OE))
		state_name = "MEMORY_RD_ASSERT_OE";		
	if ( state == (1 << MEMORY_RD_ADDRESS))
		state_name = "MEMORY_RD_ADDRESS";
	if ( state == (1 << MEMORY_RD_DATA))
		state_name = "MEMORY_RD_DATA";
	if ( state == (1 << MEMORY_RD_LATCH))
		state_name = "MEMORY_RD_LATCH";
	if ( state == (1 << MEMORY_RD_CHECK))
		state_name = "MEMORY_RD_CHECK";
	if ( state == (1 << MEMORY_RD_DELAY))
		state_name = "MEMORY_RD_DELAY";
	if ( state == (1 << START_I2C_TRANSMIT))
		state_name = "START_I2C_TRANSMIT";
	if ( state == (1 << WAIT_FOR_I2C_COMPLETE))
		state_name = "WAIT_FOR_I2C_COMPLETE";
	if ( state == (1 << CHECK_I2C_COMPLETE))
		state_name = "CHECK_I2C_COMPLETE";
	if ( state == (1 << START_I2C_TRANSMIT_DELAY))
		state_name = "START_I2C_TRANSMIT_DELAY";
	if ( state == (1 << START_I2C_RECEIVE))
		state_name = "START_I2C_RECEIVE";
	if ( state == (1 << COMPLETE_I2C_RECEIVE))
		state_name = "COMPLETE_I2C_RECEIVE";
	if ( state == (1 << MEMORY_WRITE_CYCLE))
		state_name = "MEMORY_WRITE_CYCLE";
	if ( state == (1 << MEMORY_WR_ADDRESS))
		state_name = "MEMORY_WR_ADDRESS";
	if ( state == (1 << MEMORY_WR_DATA))
		state_name = "MEMORY_WR_DATA";
	if ( state == (1 << MEMORY_WR_LATCH))
		state_name = "MEMORY_WR_LATCH";
end
`endif		
	end

//==============================================================================
//	Instantiate the Serial I2C Communications
//==============================================================================

	i2c_protocol_layer                   I2C_PROTOCOL_LAYER_INST
	(
		.CLK                            (CLK),
		.RESET                          (RST_N),

		.SERIAL_CLK                     (SERIAL_CLK),

		.I2C_DATA                       (I2C_DATA),
		.I2C_CLOCK                      (I2C_CLOCK),

		.I2C_TRANSMIT_DATA_IN           (i2c_transfer_storage_register),
		.I2C_TRANSMIT_ENA_OUT           (master_i2c_initiate_transmit),
		.I2C_TRANSMIT_RDY_IN            (master_i2c_transmit_complete),

        .I2C_TRANSMIT_DONE              (master_i2c_transmit_byte_done),		
		.I2C_RECEIVE_REQUEST            (master_i2c_initiate_receive),
		.I2C_RECEIVE_DONE               (master_i2c_receive_done),

		.I2C_RECEIVE_DATA_OUT           (serial_receive_out_byte),
		.I2C_RECEIVE_ENA_IN             (master_i2c_read_byte),
		.I2C_RECEIVE_RDY_OUT            (master_i2c_byte_received)
		);
	

endmodule

