//#######################################################################
//#
//#	Copyright 	Earth People Technology Inc. 2016
//#
//#
//# File Name:  uc_controller.v
//# Author:     Earth People Technology
//# Date:       March 5, 2016
//# Revision:   A
//#
//# Development: EPT Digital Storage Oscilloscope Project 
//# Application: Altera Cyclone IV FPGA
//# Description: This file contains verilog code which will allow access
//#              to Active Transfer Library. 
//#               
//#              
//#
//#************************************************************************
//#
//# Revision History:	
//#			DATE		VERSION		DETAILS		
//#			3/5/16	1			Created
//#
//#									
//#
//#######################################################################

`timescale 1ns/10ps

module uc_controller(


	input  wire                 CLK,
	input  wire                 RST_N,

	input  wire [7:0]           HOST_INTERFACE_ADDRESS,
	inout  wire [7:0]           HOST_INTERFACE_DATA,
	input  wire                 HOST_INTERFACE_WRITE_LATCH,
	input  wire                 HOST_INTERFACE_READ_LATCH,
	input  wire                 HOST_INTERFACE_OUTPUT_ENABLE,

	input  wire [7:0]           HOST_CONTROL_REG_NUMBER,
	inout  wire [7:0]           HOST_CONTROL_REG_CONTENTS,
	input  wire                 HOST_CONTROL_REG_NUMBER_LATCH,
	input  wire                 HOST_CONTROL_REG_CONTENTS_LATCH,
	
	output wire [7:0]           CR_0,	
	output wire [7:0]           CR_1,	
	output wire [7:0]           CR_2,	
	output wire [7:0]           CR_3,	
	output wire [7:0]           CR_4,	
	output wire [7:0]           CR_5,	
	output wire [7:0]           CR_6,	
	output wire [7:0]           CR_7,	

	input  wire [7:0]           MEMORY_INTERFACE_ADDRESS,
	inout  wire [7:0]           MEMORY_INTERFACE_DATA,
	input  wire                 MEMORY_INTERFACE_WRITE_LATCH,
	input  wire                 MEMORY_INTERFACE_READ_LATCH,
	input  wire                 MEMORY_INTERFACE_OUTPUT_ENABLE,

	input  wire                 MEMORY_INTERFACE_EN,
	output reg                  MEMORY_INTERFACE_RDY,
	
	input  wire [18:0]          TIMEOUT_TICK,
	
	output reg  [7:0]           ERROR_OUT,
	
	//output wire [11:0]          TEST_BUS_OUT,
    output reg   [3:0]          STATE_OUT


		);

//==================================================================
// Parameter Declarations
//==================================================================
	parameter	HOST_IDLE                    = 0,
                HOST_WR_ADDRESS              = 1,
                HOST_WR_DATA                 = 2,
                HOST_WR_COMPLETE             = 3,
				HOST_RD_OUTPUT_EN            = 4,
                HOST_RD_ADDRESS              = 5,
                HOST_RD_DATA                 = 6,
                HOST_RD_COMPLETE             = 7;
				
	parameter	MEMORY_IDLE                  = 0,
                MEMORY_WR_ADDRESS            = 1,
                MEMORY_WR_DATA               = 2,
                MEMORY_WR_COMPLETE           = 3,
				MEMORY_RD_OUTPUT_EN          = 4,
                MEMORY_RD_ADDRESS            = 5,
                MEMORY_RD_DATA               = 6,
                MEMORY_RD_COMPLETE           = 7;
				
	parameter	CONTROL_REG_IDLE             = 0,
                CR_NUMBER_RECIEVED           = 1,
                LATCH_CR_NUMBER              = 2,
                WAIT_FOR_CR_CONTENTS         = 3,
				LATCH_CR_CONTENTS            = 4,
                CR_COMPLETE                  = 5;
				

//==================================================================
// Register Declarations
//==================================================================
	// State Machine Registers
    reg  [7:0]                  host_state;
    reg  [7:0]                  host_next;
    reg  [7:0]                  memory_state;
    reg  [7:0]                  memory_next;
    reg  [5:0]                  control_reg_state;
    reg  [5:0]                  control_reg_next;
	
    // Local memory array
	reg     [7:0]               local_mem_array [32:0];
	
	// Host Interface Registers
    reg     [7:0]				host_address_storage;
    reg     [7:0]				host_data_storage;
    reg     [1:0]               host_mem_storage_reg;
    reg     [1:0]               host_mem_disperse_reg;

	// Memory Interface Registers Write Cycle
    reg     [7:0]               memory_address_storage;
    reg     [7:0]               memory_data_storage;
    reg     [1:0]               memory_storage_reg;
	reg                         memory_write_latch_reg;

    //Memory Interface Registers Read Cycle
    reg     [7:0]               memory_disperse;
    reg     [1:0]               memory_disperse_reg;
    reg     [7:0]               host_mem_disperse;	
	reg                         memory_read_latch;
	reg                         memory_oe_reg;
	reg                         memory_oe_flag;
	reg                         memory_read_output_enable_reg;
	reg                         memory_read_latch_reg;
	
    //Error Detection
    reg                         error_detect_reg;

    //Memory Interface Enable/Ready Handshake	
    reg     [1:0]               memory_req_grant_state;
	
    //Control Register array
	reg     [7:0]               control_reg_array [32:0];
	reg     [7:0]               control_reg_number;
	reg     [7:0]               control_reg_contents;
    reg     [1:0]               control_reg_storage_state;

//==================================================================
// Simulation Directives
//==================================================================

`ifdef SIM
	reg [8*30:1] host_state_name;
	reg [8*30:1] memory_state_name;
	reg [8*30:1] control_reg_state_name;
`endif

//=================================================================================
// Assignments	
//=================================================================================

   //Send host_mem_disperse to HOST_INTERFACE_DATA
   assign      HOST_INTERFACE_DATA = (HOST_INTERFACE_READ_LATCH & HOST_INTERFACE_OUTPUT_ENABLE ) ? host_mem_disperse : 8'hzz;

   
   //Send data out of the MEMORY_INTERFACE_DATA
   assign      MEMORY_INTERFACE_DATA = 	memory_read_latch_reg ? memory_disperse : 8'hzz;

   //Control Register Assignments
   assign      CR_0 = control_reg_array[0];
   assign      CR_1 = control_reg_array[1];
   assign      CR_2 = control_reg_array[2];
   assign      CR_3 = control_reg_array[3];
   assign      CR_4 = control_reg_array[4];
   assign      CR_5 = control_reg_array[5];
   assign      CR_6 = control_reg_array[6];
   assign      CR_7 = control_reg_array[7];

   
   //=================================================================================
   //   Register the MEMORY_INTERFACE_WRITE_LATCH
   //=================================================================================
   always @( negedge CLK or negedge RST_N )
   begin
		if( !RST_N )
		begin
		   memory_write_latch_reg <=  1'b0;
		end
		else
		begin 
		   if(MEMORY_INTERFACE_WRITE_LATCH)
              memory_write_latch_reg <= 1'b1;
		   else if(memory_state[MEMORY_WR_COMPLETE])
		      memory_write_latch_reg <= 1'b0;
        end
   end

   //=================================================================================
   //   Register the MEMORY_INTERFACE_OUTPUT_ENABLE
   //=================================================================================
   always @( negedge CLK or negedge RST_N )
   begin
		if( !RST_N )
		begin
		   memory_read_output_enable_reg <=  0;
		end
		else
		begin 
		   if(MEMORY_INTERFACE_OUTPUT_ENABLE)
              memory_read_output_enable_reg <= 1'b1;
		   else
		      memory_read_output_enable_reg <= 1'b0;
        end
   end

   //=================================================================================
   //   Register the MEMORY_INTERFACE_READ_LATCH
   //=================================================================================
   always @( negedge CLK or negedge RST_N )
   begin
		if( !RST_N )
		begin
		   memory_read_latch_reg <=  0;
		end
		else
		begin 
		   if(MEMORY_INTERFACE_READ_LATCH)
              memory_read_latch_reg <= 1'b1;
		   else
		      memory_read_latch_reg <= 1'b0;
        end
   end


//=================================================================================
   //Store the data from the MEMORY_INTERFACE_DATA
//=================================================================================
   always @( negedge CLK or negedge RST_N )
   begin
		if( !RST_N )
		begin
		   memory_data_storage <=  0;
		end
		else
		begin
           if(MEMORY_INTERFACE_READ_LATCH)
              memory_data_storage <=  MEMORY_INTERFACE_DATA;
           else if(MEMORY_INTERFACE_WRITE_LATCH)
              memory_data_storage <=  MEMORY_INTERFACE_DATA;
        end
   end
   
   
//=================================================================================
//	Control Register Store Contents 
//=================================================================================

	always @( negedge CLK or negedge RST_N )
	begin
		if( !RST_N )
		begin
		   control_reg_number <= 0;
		   control_reg_contents <= 0;
		   control_reg_storage_state <= 2'b00;
		end
		else
		begin
		   case(control_reg_storage_state)
		   2'b00:
		   begin
		      if(HOST_CONTROL_REG_NUMBER_LATCH)
			  begin
		         control_reg_number <= HOST_CONTROL_REG_NUMBER;
		         control_reg_storage_state <= 2'b01;
			  end
		   end
		   2'b01:
		   begin
		      if(HOST_CONTROL_REG_CONTENTS_LATCH)
			  begin
                 control_reg_contents <= HOST_CONTROL_REG_CONTENTS;
				 control_reg_storage_state <= 2'b10;
			  end
		   end
		   2'b10:
		   begin
		         control_reg_array[control_reg_number] <= control_reg_contents;
		         control_reg_storage_state <= 2'b11;
		   end
		   2'b11:
		   begin
		         control_reg_storage_state <= 2'b00;
		   end
		   endcase
		   
		   //Self Resetting register bits
		   if(TIMEOUT_TICK[10])
		      control_reg_array[5] <= 0;
		end
	end

//=================================================================================
//  Local Memory Controller
//=================================================================================

	always @( negedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		begin
		end
		else
		begin
            if(host_mem_storage_reg[1])
			begin
			   local_mem_array[host_address_storage] <= host_data_storage;
			end
			else if(memory_storage_reg[1])
			begin
			   local_mem_array[memory_address_storage] <= memory_data_storage;
			end
			else if(host_mem_disperse_reg[1])
			begin
			    host_mem_disperse <= local_mem_array[host_address_storage];
			end
			else if(memory_disperse_reg[1])
			begin
			    memory_disperse <= local_mem_array[memory_address_storage];
			end
	    end
	end
   

   //-----------------------------------------------
   // Create a State Test Output Signal
   //           MEMORY_IDLE                  = 0,
   //           MEMORY_WR_ADDRESS            = 1,
   //           MEMORY_WR_DATA               = 2,
   //           MEMORY_WR_COMPLETE           = 3,
   //			MEMORY_RD_OUTPUT_EN          = 4,
   //           MEMORY_RD_ADDRESS            = 5,
   //           MEMORY_RD_DATA               = 6,
   //           MEMORY_RD_COMPLETE           = 7;
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	  STATE_OUT <= 0;
	else
	  begin
	     if (memory_state[MEMORY_IDLE])
           STATE_OUT <= 5'h0;
	     if (memory_state[MEMORY_WR_ADDRESS])
	       STATE_OUT <= 5'h1;
	     if (memory_state[MEMORY_WR_DATA])
	       STATE_OUT <= 5'h2;
	     if (memory_state[MEMORY_WR_COMPLETE])
	       STATE_OUT <= 5'h3;
	     if (memory_state[MEMORY_RD_OUTPUT_EN])
	       STATE_OUT <= 5'h4;
	     if (memory_state[MEMORY_RD_ADDRESS])
	       STATE_OUT <= 5'h5;
	     if (memory_state[MEMORY_RD_DATA])
	       STATE_OUT <= 5'h6;
	     if (memory_state[MEMORY_RD_COMPLETE])
	       STATE_OUT <= 5'h7;
	  end // else: !if(!RST_N)
     end // always @ (posedge CLK or negedge RST_N)


	
//=================================================================================
//	ERROR Detection: Error Register captures collisions between the Host 
//  Interface and the Memory Interface
//=================================================================================

	always @( negedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		begin
			ERROR_OUT <= 0;			
			error_detect_reg <= 1'b0;
		end
		else
		begin
		    if((HOST_INTERFACE_WRITE_LATCH & (memory_write_latch_reg | memory_read_latch_reg)) & !error_detect_reg)
			begin
			   ERROR_OUT <= 8'h01;			
			   error_detect_reg <= 1'b1;
			end
			else if((HOST_INTERFACE_READ_LATCH & (memory_write_latch_reg | memory_read_latch_reg)) & !error_detect_reg)
			begin
			   ERROR_OUT <= 8'h02;
			   error_detect_reg <= 1'b0;
			end
		    else if(host_state[HOST_IDLE])
			begin
			   ERROR_OUT <= 8'h00;
			   error_detect_reg <= 1'b0;
			end
		end
	end

	
//=================================================================================
//	Memory Read/Write Cycle: Enable and Ready Request of Local Memory
//=================================================================================

	always @( negedge CLK or negedge RST_N )
	begin
		if( !RST_N )
		begin
		      memory_req_grant_state <= 2'b00;
			  MEMORY_INTERFACE_RDY <= 1'b0;
		end
		else
		begin
		   case(memory_req_grant_state)
		   2'b00:
		   begin
		      if(MEMORY_INTERFACE_EN)
		         memory_req_grant_state <= 2'b01;
		   end
		   2'b01:
		   begin
		       if(!host_mem_storage_reg[1] & !host_mem_disperse_reg[1])
			   begin
                   MEMORY_INTERFACE_RDY <= 1'b1;
				   memory_req_grant_state <= 2'b10;
			   end
			   else if(!MEMORY_INTERFACE_EN)
				   memory_req_grant_state <= 2'b00;
		   end
		   2'b10:
		   begin
		      if(memory_state[MEMORY_IDLE] & !MEMORY_INTERFACE_EN)
			  begin
                 MEMORY_INTERFACE_RDY <= 1'b0;
		         memory_req_grant_state <= 2'b00;
			  end
		      if(memory_state[MEMORY_WR_DATA])
			  begin
                 MEMORY_INTERFACE_RDY <= 1'b0;
		         memory_req_grant_state <= 2'b00;
			  end
		   end
		   endcase
		end
	end
	


//=================================================================================
//	Memory Write Cycle: 
//=================================================================================

	always @( negedge CLK or negedge RST_N )
	begin
		if( !RST_N )
		begin
		      memory_storage_reg <= 2'b00;
		end
		else
		begin
		   if(memory_write_latch_reg & (memory_storage_reg == 2'b00))
		   begin
		      memory_storage_reg <= 2'b10;
		   end
		   else if(memory_state[MEMORY_WR_COMPLETE] & (memory_storage_reg == 2'b10))
		   begin
		      memory_storage_reg <= 2'b11;
		   end
		   else if(memory_state[MEMORY_IDLE] & (memory_storage_reg == 2'b11))
		   begin
		      memory_storage_reg <= 2'b00;
		   end
		end
	end
	

//=================================================================================
//	Memory Write/Read Cycle: Memory Address Storage
//=================================================================================

	always @( negedge CLK or negedge RST_N )
	begin
		if( !RST_N )
		begin
		      memory_address_storage <= 0;
		      memory_disperse_reg <= 2'b00;
		end
		else
		begin
		   if(memory_state[MEMORY_RD_ADDRESS]  & (memory_disperse_reg == 2'b00))
		   begin
		      memory_address_storage <= MEMORY_INTERFACE_ADDRESS;
		      memory_disperse_reg <= 2'b01;
		   end
		   else if((memory_state[MEMORY_RD_DATA] | memory_state[MEMORY_WR_DATA]) & (memory_disperse_reg == 2'b01))
		   begin
		      memory_disperse_reg <= 2'b10;
		   end
		   else if((memory_state[MEMORY_RD_COMPLETE] | memory_state[MEMORY_WR_COMPLETE]) & (memory_disperse_reg == 2'b10))
		   begin
		      memory_disperse_reg <= 2'b11;
		   end
		   if(MEMORY_INTERFACE_WRITE_LATCH)
		   begin
		      memory_address_storage <= MEMORY_INTERFACE_ADDRESS;
		   end
		   else if(memory_state[MEMORY_IDLE] & (memory_disperse_reg == 2'b11))
		   begin
		      memory_disperse_reg <= 2'b00;
		   end
		end
	end

//=================================================================================
//	Memory Read Cycle: Register the MEMORY_INTERFACE_READ_LATCH
//=================================================================================

	always @( negedge CLK or negedge RST_N )
	begin
		if( !RST_N )
		begin
		      memory_read_latch <= 1'b0;
		end
		else
		begin
		   if(memory_read_latch_reg)
		   begin
		      memory_read_latch <= 1'b1;
		   end
		   else if(memory_next[MEMORY_RD_COMPLETE])
		   begin
		      memory_read_latch <= 1'b0;
		   end
		end
	end

//=================================================================================
//	Memory Read Cycle: Register the MEMORY_INTERFACE_OUTPUT_ENABLE
//=================================================================================

	always @( negedge CLK or negedge RST_N )
	begin
		if( !RST_N )
		begin
		      memory_oe_flag <= 1'b0;
		      memory_oe_reg <= 1'b0;
		end
		else
		begin
		   if(memory_read_output_enable_reg & !memory_oe_reg)
		   begin
		      memory_oe_flag <= 1'b1;
		      memory_oe_reg <= 1'b1;
		   end
		   else if(memory_state[MEMORY_RD_COMPLETE])
		   begin
		      memory_oe_flag <= 1'b0;
		      memory_oe_reg <= 1'b1;
		   end
		   else if(memory_state[MEMORY_IDLE])
		   begin
		      memory_oe_flag <= 1'b0;
		      memory_oe_reg <= 1'b0;
		   end
		end
	end


 //=================================================================================
 //	Host Write Cycle: Store the Address and Data from the Host
 //=================================================================================


	always @( negedge CLK or negedge RST_N )
	begin
		if( !RST_N )
		begin
           host_address_storage <= 0;
           host_data_storage <= 0;
		end
		else
		begin
		   if(host_state[HOST_WR_ADDRESS] | host_state[HOST_RD_ADDRESS])
		      host_address_storage <= HOST_INTERFACE_ADDRESS;
		   if(host_state[HOST_WR_DATA])
		      host_data_storage <= HOST_INTERFACE_DATA;
		end
	end

 //=================================================================================
 //	Host Write Cycle: Store Data from the Host into Memory
 //=================================================================================


	always @( negedge CLK or negedge RST_N )
	begin
		if( !RST_N )
		begin
		      host_mem_storage_reg <= 2'b00;
		end
		else
		begin
		   if(host_state[HOST_WR_COMPLETE] & (host_mem_storage_reg == 2'b00))
		   begin
		      host_mem_storage_reg <= 2'b11;
		   end
		   else if(host_state[HOST_WR_COMPLETE] & (host_mem_storage_reg == 2'b11))
		   begin
		      host_mem_storage_reg <= 2'b01;
		   end
		   else if(host_state[HOST_IDLE] & (host_mem_storage_reg == 2'b01))
		   begin
		      host_mem_storage_reg <= 2'b00;
		   end
		end
	end



//=================================================================================
//	Host Read Cycle: Send local memory data byte to HOST_INTERFACE_DATA
//=================================================================================


	always @( negedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		begin
		      host_mem_disperse_reg <= 2'b00;
		end
		else
		begin
		   if(host_state[HOST_RD_DATA] & (host_mem_disperse_reg == 2'b00))
		   begin
		      host_mem_disperse_reg <= 2'b11;
		   end
		   else if(host_state[HOST_RD_COMPLETE] & (host_mem_disperse_reg == 2'b11))
		   begin
		      host_mem_disperse_reg <= 2'b01;
		   end
		   else if(host_state[HOST_IDLE] & (host_mem_disperse_reg == 2'b01))
		   begin
		      host_mem_disperse_reg <= 2'b00;
		   end
		end
	end
		
	
//=================================================================================
//	Host Bus Communications Finite host_state Machine
//
//           	HOST_IDLE                    = 0,
//              HOST_WR_ADDRESS              = 1,
//              HOST_WR_DATA                 = 2,
//              HOST_WR_COMPLETE             = 3,
//				HOST_RD_OUTPUT_EN            = 4,
//              HOST_RD_ADDRESS              = 5,
//              HOST_RD_DATA                 = 6,
//              HOST_RD_COMPLETE             = 7;
//
//
//=================================================================================

	always @(  posedge CLK or negedge RST_N  )
	begin
		if( !RST_N )
		begin
			host_state <= 8'h00;
			host_state [HOST_IDLE] <= 1'b1;
		end
		else
			host_state <= host_next;
	end


	always @( host_state or HOST_INTERFACE_WRITE_LATCH or HOST_INTERFACE_OUTPUT_ENABLE 
	        or HOST_INTERFACE_READ_LATCH
		    )
	begin
		host_next = 8'h00;
		if (host_state[HOST_IDLE])
		begin
			if (HOST_INTERFACE_WRITE_LATCH)
				host_next[HOST_WR_ADDRESS] = 1'b1;
		    else if(HOST_INTERFACE_OUTPUT_ENABLE)
				host_next[HOST_RD_OUTPUT_EN] = 1'b1;
			else
				host_next[HOST_IDLE] = 1'b1;
		end

		if (host_state[HOST_WR_ADDRESS])
		begin
		   host_next[HOST_WR_DATA] = 1'b1;
		end

		if (host_state[HOST_WR_DATA])
		begin
		   host_next[HOST_WR_COMPLETE] = 1'b1;
		end

		if (host_state[HOST_WR_COMPLETE])
		begin
		    if(!HOST_INTERFACE_WRITE_LATCH)
				host_next[HOST_IDLE] = 1'b1;
			else
				host_next[HOST_WR_COMPLETE] = 1'b1;
		end

		if (host_state[HOST_RD_OUTPUT_EN])
		begin
		   host_next[HOST_RD_ADDRESS] = 1'b1;
		end

		if (host_state[HOST_RD_ADDRESS])
		begin
		   host_next[HOST_RD_DATA] = 1'b1;
		end

		if (host_state[HOST_RD_DATA])
		begin
			if (HOST_INTERFACE_READ_LATCH)
		       host_next[HOST_RD_COMPLETE] = 1'b1;
			else
				host_next[HOST_RD_DATA] = 1'b1;
		end

		if (host_state[HOST_RD_COMPLETE])
		begin
			if (!HOST_INTERFACE_READ_LATCH)
				host_next[HOST_IDLE] = 1'b1;
			else
				host_next[HOST_RD_COMPLETE] = 1'b1;
		end


//=================================================================
// For Simumlation, display the host_state names
//
//=================================================================

`ifdef SIM
	if ( host_state == (1 << HOST_IDLE))
		host_state_name = "HOST_IDLE";		
	if ( host_state == (1 << HOST_WR_ADDRESS))
		host_state_name = "HOST_WR_ADDRESS";		
	if ( host_state == (1 << HOST_WR_DATA))
		host_state_name = "HOST_WR_DATA"; 		
	if ( host_state == (1 << HOST_WR_COMPLETE))
		host_state_name = "HOST_WR_COMPLETE";		
	if ( host_state == (1 << HOST_RD_OUTPUT_EN))
		host_state_name = "HOST_RD_OUTPUT_EN";		
	if ( host_state == (1 << HOST_RD_ADDRESS))
		host_state_name = "HOST_RD_ADDRESS";		
	if ( host_state == (1 << HOST_RD_DATA))
		host_state_name = "HOST_RD_DATA";
	if ( host_state == (1 << HOST_RD_COMPLETE))
		host_state_name = "HOST_RD_COMPLETE";
//end
`endif		
	end

//=================================================================================
//	Memory Bus Communications Finite host_state Machine
//
//
/*				
	parameter	MEMORY_IDLE                  = 0,
                MEMORY_WR_ADDRESS            = 1,
                MEMORY_WR_DATA               = 2,
                MEMORY_WR_COMPLETE           = 3,
				MEMORY_RD_OUTPUT_EN          = 4,
                MEMORY_RD_ADDRESS            = 5,
                MEMORY_RD_DATA               = 6,
                MEMORY_RD_COMPLETE           = 7;
				
*///
//=================================================================================

	always @(  posedge CLK or negedge RST_N  )
	begin
		if( !RST_N )
		begin
			memory_state <= 8'h00;
			memory_state [MEMORY_IDLE] <= 1'b1;
		end
		else
			memory_state <= memory_next;
	end


	always @( memory_state or memory_write_latch_reg  
	        or memory_read_output_enable_reg or memory_read_latch
		    or memory_read_latch_reg or memory_oe_flag)
	begin
		memory_next = 8'h00;
		if (memory_state[MEMORY_IDLE])
		begin
			if (memory_write_latch_reg)
				memory_next[MEMORY_WR_ADDRESS] = 1'b1;
		    else if(memory_read_output_enable_reg)
				memory_next[MEMORY_RD_OUTPUT_EN] = 1'b1;
			else
				memory_next[MEMORY_IDLE] = 1'b1;
		end

		if (memory_state[MEMORY_WR_ADDRESS])
		begin
		   memory_next[MEMORY_WR_DATA] = 1'b1;
		end

		if (memory_state[MEMORY_WR_DATA])
		begin
		    if(memory_write_latch_reg)
		       memory_next[MEMORY_WR_COMPLETE] = 1'b1;
			else
		       memory_next[MEMORY_WR_DATA] = 1'b1;
		end

		if (memory_state[MEMORY_WR_COMPLETE])
		begin
		    if(!memory_write_latch_reg)
				memory_next[MEMORY_IDLE] = 1'b1;
			else
				memory_next[MEMORY_WR_COMPLETE] = 1'b1;
		end

		if (memory_state[MEMORY_RD_OUTPUT_EN])
		begin
		   memory_next[MEMORY_RD_ADDRESS] = 1'b1;
		end

		if (memory_state[MEMORY_RD_ADDRESS])
		begin
		   memory_next[MEMORY_RD_DATA] = 1'b1;
		end

		if (memory_state[MEMORY_RD_DATA])
		begin
			if (memory_read_latch_reg)
		       memory_next[MEMORY_RD_COMPLETE] = 1'b1;
			else
				memory_next[MEMORY_RD_DATA] = 1'b1;
		end

		if (memory_state[MEMORY_RD_COMPLETE])
		begin
			if (!memory_read_latch & !memory_oe_flag)
				memory_next[MEMORY_IDLE] = 1'b1;
			else
				memory_next[MEMORY_RD_COMPLETE] = 1'b1;
		end


//=================================================================
// For Simumlation, display the host_state names
//
//                MEMORY_IDLE                  = 0,
//                MEMORY_WR_ADDRESS            = 1,
//                MEMORY_WR_DATA               = 2,
//                MEMORY_WR_COMPLETE           = 3,
//				  MEMORY_RD_OUTPUT_EN          = 4,
//                MEMORY_RD_ADDRESS            = 5,
//                MEMORY_RD_DATA               = 6,
//                MEMORY_RD_COMPLETE           = 7;
//				
//
//=================================================================

`ifdef SIM
	if ( memory_state == (1 << MEMORY_IDLE))
		memory_state_name = "MEMORY_IDLE";		
	if ( memory_state == (1 << MEMORY_WR_ADDRESS))
		memory_state_name = "MEMORY_WR_ADDRESS";		
	if ( memory_state == (1 << MEMORY_WR_DATA))
		memory_state_name = "MEMORY_WR_DATA"; 		
	if ( memory_state == (1 << MEMORY_WR_COMPLETE))
		memory_state_name = "MEMORY_WR_COMPLETE";		
	if ( memory_state == (1 << MEMORY_RD_OUTPUT_EN))
		memory_state_name = "MEMORY_RD_OUTPUT_EN";		
	if ( memory_state == (1 << MEMORY_RD_ADDRESS))
		memory_state_name = "MEMORY_RD_ADDRESS";		
	if ( memory_state == (1 << MEMORY_RD_DATA))
		memory_state_name = "MEMORY_RD_DATA";
	if ( memory_state == (1 << MEMORY_RD_COMPLETE))
		memory_state_name = "MEMORY_RD_COMPLETE";
//end
`endif		
	end

//=================================================================================
//	Control Register Finite host_state Machine
//
//
//
//=================================================================================

	always @(  posedge CLK or negedge RST_N  )
	begin
		if( !RST_N )
		begin
			control_reg_state <= 6'h00;
			control_reg_state [CONTROL_REG_IDLE] <= 1'b1;
		end
		else
			control_reg_state <= control_reg_next;
	end


	always @( control_reg_state or HOST_CONTROL_REG_NUMBER_LATCH or HOST_CONTROL_REG_CONTENTS_LATCH 
	        or memory_read_latch_reg
		    )
	begin
		control_reg_next = 6'h00;
		if (control_reg_state[CONTROL_REG_IDLE])
		begin
			if (HOST_CONTROL_REG_NUMBER_LATCH)
				control_reg_next[CR_NUMBER_RECIEVED] = 1'b1;
			else
				control_reg_next[CONTROL_REG_IDLE] = 1'b1;
		end

		if (control_reg_state[CR_NUMBER_RECIEVED])
		begin
		   control_reg_next[LATCH_CR_NUMBER] = 1'b1;
		end

		if (control_reg_state[LATCH_CR_NUMBER])
		begin
		   control_reg_next[WAIT_FOR_CR_CONTENTS] = 1'b1;
		end

		if (control_reg_state[WAIT_FOR_CR_CONTENTS])
		begin
		    if(HOST_CONTROL_REG_CONTENTS_LATCH)
				control_reg_next[LATCH_CR_CONTENTS] = 1'b1;
			else
				control_reg_next[WAIT_FOR_CR_CONTENTS] = 1'b1;
		end

		if (control_reg_state[LATCH_CR_CONTENTS])
		begin
		   control_reg_next[CR_COMPLETE] = 1'b1;
		end

		if (control_reg_state[CR_COMPLETE])
		begin
		    if(!HOST_CONTROL_REG_CONTENTS_LATCH)
				control_reg_next[CONTROL_REG_IDLE] = 1'b1;
			else
				control_reg_next[CR_COMPLETE] = 1'b1;
		end


//=================================================================
// For Simumlation, display the host_state names
//
//                CONTROL_REG_IDLE       = 0,
//                CR_NUMBER_RECIEVED     = 1,
//                LATCH_CR_NUMBER        = 2,
//                WAIT_FOR_CR_CONTENTS   = 3,
//				  LATCH_CR_CONTENTS      = 4,
//                CR_COMPLETE            = 5,
//				
//
//=================================================================

`ifdef SIM
	if ( control_reg_state == (1 << CONTROL_REG_IDLE))
		control_reg_state_name = "CONTROL_REG_IDLE";		
	if ( control_reg_state == (1 << CR_NUMBER_RECIEVED))
		control_reg_state_name = "CR_NUMBER_RECIEVED";		
	if ( control_reg_state == (1 << LATCH_CR_NUMBER))
		control_reg_state_name = "LATCH_CR_NUMBER"; 		
	if ( control_reg_state == (1 << WAIT_FOR_CR_CONTENTS))
		control_reg_state_name = "WAIT_FOR_CR_CONTENTS";		
	if ( control_reg_state == (1 << LATCH_CR_CONTENTS))
		control_reg_state_name = "LATCH_CR_CONTENTS";		
	if ( control_reg_state == (1 << CR_COMPLETE))
		control_reg_state_name = "CR_COMPLETE";		
//end
`endif		
	end
endmodule

